/* uzel.c */

/* Do komunikačního uzlu počítačové sítě přichází pakety v Poissonovském toku
 * (intenzita LAMBDA). Komunikační uzel část paketů (K procent) neposílá dál
 * (řídící pakety, doba zpracování je T1 plus/minus DELTA - rovnoměrné
 * rozdělení) a zbytek paketů odesílá dál. Odesílaných (datových) paketů jsou
 * dva typy (náhodně smíchané a rovnoměrně zastoupené ve vstupním proudu) doba
 * obsluhy je nenáhodná (T2 pro první typ, T3 pro druhý typ). Chceme simulací
 * určit střední zpoždění datových paketů v uzlu (tj. střední dobu jejich
 * průchodu uzlem).
 */


#include <stdio.h>
#include <math.h>

#include <csim/csim.h>
#include <csim/csim_rng.h>
#include <csim/csim_err.h>


#define LAMBDA 0.5 /* střední rychlost vstupujících paketů */
#define K 0.4      /* poměr řídících a datových paketů  */
#define T1 1.0     /* střední doba zpracování řídícího paketu */
#define DELTA 0.1  /* rozsah v rovnoměrném rozdělení t1 */
#define T2 1.5     /* střední doba obsluhy prvního typu datových paketů */
#define T3 2.5     /* střední doba obsluhy druhého typu datových paketů */

#define MAX_FIFO_LEN 5  /* maximální evidovaná délka fronty */

/* paket */
typedef struct {
  CSIM_LINK link;       /* typ LINK */
  double start;         /* čas vzniku požadavku */
  int management;       /* zda je řídící */
  int type;             /* typ datového paketu */
} PACKET;


typedef struct {
  CSIM_HEAD head;
  double last_time;    /* čas posledních statistik fronty */
  double p[MAX_FIFO_LEN + 1];  /* délky front vážené časem */
} FIFO;

/* uzel */
typedef struct {
  CSIM_PROCESS process; /* typ PROCESS */
  FIFO *fifo;           /* odkaz na frontu (odebírání paketů) */
  /* statistiky */
  double sum_time;      /* celkový čas všech požadavků */
  double sum_time2;     /* součt čtverců časů požadavků */
  unsigned long count;  /* celkový počet požadavků */
} NODE;

/* generátor (vstup) */
typedef struct {
  CSIM_PROCESS process; /* typ PROCESS */
  FIFO *fifo;           /* odkaz na frontu (ukládání paketů) */
  NODE *node;           /* odkaz na uzel (probouzení po vložení paketu) */
} INPUT;

CSIM_RNG_STATE *rng;   /* objekt generátoru */


/* 
 * aktualizuje statistiky související s frontou 
 * (volat vždy po práci s frontou)
 */
static void fifo_update(FIFO *fifo) {
  size_t index;

  index = csim_cardinal((CSIM_HEAD *)fifo);
  if (index > MAX_FIFO_LEN) index = MAX_FIFO_LEN;
  fifo->p[index] += (csim_time() - fifo->last_time);
  fifo->last_time = csim_time();
}


/* program vstupu */
csim_program(INPUT, PROG_INPUT)
  PACKET *packet;

  while (1) {
    packet = csim_new_link(PACKET);
    packet->start = csim_time();
    packet->management = csim_draw(rng, K);
    if (!packet->management) packet->type = csim_draw(rng, 0.5);
    else packet->type = -1;
    csim_into((CSIM_LINK *)packet, (CSIM_HEAD *)my.fifo);
    fifo_update(my.fifo);
    if (csim_idle((CSIM_PROCESS *)my.node)) csim_activate_at((CSIM_PROCESS *)my.node, csim_time());
    csim_hold(csim_negexp(rng, LAMBDA));
  }
csim_end_program


/* program procesu uzlu */
csim_program(NODE, PROG_NODE)
  PACKET *packet;
  double time;

  while (1) {
    /* čekáme na nějaká paket */
    while (csim_empty((CSIM_HEAD *)my.fifo)) csim_passivate();
    /* první čekající požadavek z fronty */
    packet = (PACKET *)csim_first((CSIM_HEAD *)my.fifo);
    /* akce a výpočet času čekání */
    if (packet->management) time = csim_uniform(rng, T1 - DELTA, T1 + DELTA);
    else {
      /* datový paket, děláme statistiky */
      time = packet->type == 0 ? T1 : T2;
      my.sum_time += time;
      my.sum_time2 += time * time;
      my.count++;
    }
    csim_out((CSIM_LINK *)packet);
    csim_dispose_link((CSIM_LINK *)packet);
    fifo_update(my.fifo);
    csim_hold(time);
  }
csim_end_program


void init(INPUT **input, FIFO **fifo, NODE **node) {
  int i;

  /* fronta mezi vstupem a uzlem */
  *fifo = csim_new_head(FIFO);
  (*fifo)->last_time = 0.0;
  for (i = 0; i <= MAX_FIFO_LEN; i++) (*fifo)->p[i] = 0.0;

  /* objekt vstupu */
  *input = csim_new_process(INPUT, PROG_INPUT);

  /* objekt uzlu */
  *node = csim_new_process(NODE, PROG_NODE);
  (*node)->sum_time = 0;
  (*node)->sum_time2 = 0;
  (*node)->count = 0;

  /* vazby objektů */
  (*input)->node = *node;
  (*input)->fifo = *fifo;
  (*node)->fifo = *fifo;

  /* aktivace */
  csim_activate_at((CSIM_PROCESS *)*input, 0.0);
  csim_activate_at((CSIM_PROCESS *)*node, 0.0);
}


int main() {
  INPUT *input;       /* vstup */
  FIFO *fifo;         /* fronta */
  NODE *node;         /* uzel */
  CSIM_ERROR err;
  long int cnt, limit, cnt_max = 50000;
  int ui_cnt, i;
  double Ex, Ex2, Dx; /* vysledky */

  csim_init_mem();
  rng = csim_new_rng(INIT_SEED);

  init(&input, &fifo, &node);

  printf("Model komunikacniho uzlu:\n");
  printf("  stredni vstupni frekvence: LAMBDA = %.2f\n", LAMBDA);
  printf("  pomer ridicich a datovych paketu: K = %.2f %%\n", 100 * K);
  printf("  doba zpracovani ridiciho paketu: t1 = %.2f +- %.2f\n", T1, DELTA);
  printf("  stredni doba zpracovani datoveho paketu 1. typu: T2 = %.2f\n", T2);
  printf("  stredni doba zpracovani datoveho paketu 2. typu: T3 = %.2f\n", T3);
  printf("\n");

  printf("Simulace:\n");
  cnt = 0;
  ui_cnt = 0;
  limit = 0;
  while (cnt < cnt_max) {
    if (csim_step() == FAILURE) {
      csim_error(&err);
      printf("%s\n", csim_error_msg(err.error_code));
      break;
    }
    if (cnt++ >= limit) {
      if (++ui_cnt % 5 == 0) printf("%d%%", 2*ui_cnt);
      else printf(".");
      fflush(stdout);
      limit = cnt_max * ui_cnt / 50;
    }
  }
  printf("\n\n");

  Ex = node->sum_time / node->count;
  Ex2 = node->sum_time2 / node->count;
  Dx = Ex2 - Ex * Ex;
  for (i = 0; i <= MAX_FIFO_LEN; i++) fifo->p[i] /= csim_time();
  printf("Simulovana doba: %lu\n", cnt);
  printf("Celkovy pocet datovych paketu: %lu\n", node->count);
  printf("Prumerna doba zpracovani datoveho paketu: %f\n", Ex);
  printf("Prumerna smerodatna odchylka doby zpracovani paketu: %f\n", sqrt(Dx));
  printf("Pravdepodobnosti delek front:\n");
  for (i = 0; i < MAX_FIFO_LEN; i++) {
    printf("  p(lw = %d) = %f\n", i, fifo->p[i]);
  }
  printf("  p(lw > %d) = %f\n", MAX_FIFO_LEN, fifo->p[MAX_FIFO_LEN]);

  csim_clear_mem();

  return 0;
}
